
function cgilua.chunk( l )
  local chunk
  local first = strsub( l.chunk, 1, 1 )
  if first == '@' then
    chunk = format( " at <i>%s</i>", strsub(l.chunk, 2) )
  elseif first == '(' then
    chunk = " at C code"
  else

    -- check if chunk is an embedded code
    local emb = cgilua.embedded[l.chunk]
    if emb then

      if l.line > 0 then
        l.line = l.line + emb[1] - 1
        chunk = format( " at <i>%s</i>", emb[2] ) 
      else
        chunk = format( " starting at line %d, file <i>%s</i>", emb[1], emb[2] ) 
      end

    else

      local str = l.chunk
      str = gsub( str, "%s%s*", " " );
      str = gsub( str, "^ ", "" );
      if strlen(str) > 50 then
       str = strsub( str, 1, 50 ) .. ' ...'
      end
      str = gsub( str, "<", "&lt;" )
      str = gsub( str, " ", "&nbsp;" )
      chunk = format( ' at string "<code>%s</code>"', str )

    end
  end
  if l.line > 0 then
    chunk = chunk .. format( ", line %d", l.line ) 
  end
  return chunk
end

-- <cgilua.stacklevel> returns one level of the stack
-- described in <l> in a string.

function cgilua.stacklevel( l )
  local level
  if l.type == 'global' then
    level = format( "function <code>%s</code>", l.name )
  elseif l.type == 'tag-method' then
    level = format( "<code>%s</code> tag method", l.name )
  elseif l.func and cgilua.functionname[l.func] then
    level = format( "function <code>cgilua.%s</code>", 
            cgilua.functionname[l.func] )
  elseif l.type ~= '' then
    level = l.type
  else
    if l.linedef == 0 then
      level = "user script"
    else
      level = gsub(tostring(l.func),":","")
    end
  end
  return level .. cgilua.chunk( l )
end


-- this function returns a modified Lua stack, hiding 
-- some internal processing and showing in the stack
-- embedded code, html files etc.
-- It searches the stack for the functions cgilua.includehtml, cgilua.usererror,
-- cgilua.run, cgilua.doscript, cgilua.douserfile and cgilua.douserstring.

function cgilua.getstack()
  local stack = cgilua.getluastack() -- lua stack
  local userstack = {}  -- final table to be printed
  local i = 4         -- index for the 'stack' table (ignore the top 3 functions)
  local j = 1         -- index for the 'userstack' table

  while stack[i] do

    -- cgilua.includehtml:
    -- remove it from the stack and put the name of the
    -- html file being processed in the same way as a lua file
    -- executed with 'dofile'
    if stack[i].func == cgilua.includehtml then

      local str = cgilua.pop( 'script' )
      if str ~= '' then
        userstack[j] = {}
        userstack[j].type = ''
        userstack[j].line = 0
        userstack[j].linedef = 0
        userstack[j].chunk = '@'..str
        j = j + 1
      end

      userstack[j] = stack[i]
      i = i + 1
      j = j + 1
      
    -- cgilua.run:
    -- this is the base of the stack.
    -- it calls cgilua.doscript, which is removed from the stack.
    -- this function and the following stack levels are also removed
    elseif stack[i].func == cgilua.run then

      userstack[j-1] = nil
      stack[i] = nil

    -- cgilua.douserfile:
    -- remove it from the stack ('dofile' is enough)
    elseif stack[i].func == cgilua.douserfile then

      j = j - 1
      userstack[j] = nil
      i = i + 1

    -- cgilua.usererror:
    -- remove it and the pre-process functions following this stack level
    elseif stack[i].func == cgilua.usererror then
    
      j = j - 1
      userstack[j] = nil
      i = i + 1
      while stack[i] and
            stack[i].func ~= cgilua.pre_ProcessFile and
            cgilua.prefunction[stack[i].func] do
        i = i + 1
      end

    -- cgilua.douserstring:
    -- shows the dostring as an embedded code and remove
    -- the pre-process functions following this stack level
    elseif stack[i].func == cgilua.douserstring then

      if j > 2 then
        j = j - 1
        userstack[j] = nil
      end

      local tab = cgilua.pop( 'embedded' )
      userstack[j-1].chunk = tab[1]
      userstack[j-1].type = tab[2]

      i = i + 1
      while stack[i] and
            stack[i].func ~= cgilua.pre_ProcessFile and
            cgilua.prefunction[stack[i].func] do
        i = i + 1
      end

    else
      userstack[j] = stack[i]
      i = i + 1
      j = j + 1
    end
  end

  local stacklevel = {}
  i = 1
  while userstack[i] do
    stacklevel[i] = cgilua.stacklevel( userstack[i] )
    i = i + 1
  end
  return stacklevel
end


-- this is the function called by the error fallback
-- to show the Lua stack. 

function cgilua.printstack()
  local toprint = cgilua.getstack()
  i = 1
  while toprint[i] do
    write( toprint[i] .. '<br>\n' )
    i = i + 1
  end
end


-- the functions <cgilua.{create_stack, push, pop}>
-- are used to manage the cgilua stacks

cgilua.stacks = {}
cgilua.stacks.tops = {}

function cgilua.create_stack( name )
  cgilua.stacks[name] = {}
  cgilua.stacks.tops[name] = 0
end

function cgilua.push( stack, value )
  local top = cgilua.stacks.tops[stack]
  cgilua.stacks[stack][ top ] = value
  cgilua.stacks.tops[stack] = top + 1
end

function cgilua.pop( stack )
  local top = cgilua.stacks.tops[stack] - 1
  cgilua.stacks.tops[stack] = top 
  return cgilua.stacks[stack][ top ]
end


-- create two stacks

cgilua.create_stack( 'script' )
cgilua.create_stack( 'embedded' )
cgilua.embedded = {}

-- this function can be used to define an user error method
-- ths error method default (cgilua.usererrorfallback) shows
-- the stack as an html page.

function cgilua.seterrormethod( fn )
  cgilua.userdefinederrormethod = fn
end

cgilua.seterrormethod( cgilua.usererrormethod )

-- this is the real error method set in lua before the
-- execution of an user code.

function cgilua.internalerrormethod( msg )
  seterrormethod( cgilua.usererrormethod )
  cgilua.userdefinederrormethod( msg )
  exit()
end 

-- define wrappers for the functions dofile and dostring,
-- setting the user_fallback before the call to dofile (dostring)

function cgilua.douserstring( luacode, embedded, line, file )
  if line then
    cgilua.embedded[luacode] = { line, file }
  end
  cgilua.push( 'embedded', { luacode, embedded } )
  local old = seterrormethod( cgilua.internalerrormethod );
  result = dostring( luacode )
  seterrormethod( old );
  cgilua.pop( 'embedded' )
  return result
end

function cgilua.douserfile( file )
  local old = seterrormethod( cgilua.internalerrormethod )  
  local result = dofile( file )
  seterrormethod( old )
  return result
end

function cgilua.usererror( msg )
  seterrormethod( cgilua.internalerrormethod )  
  error( msg )
end

cgilua.prefunction = {}
cgilua.prefunction[ cgilua.pre_ReplaceTag ] = 1
cgilua.prefunction[ cgilua.pre_ProcessTags] = 1
cgilua.prefunction[ cgilua.pre_PrepareText] = 1
cgilua.prefunction[ cgilua.pre_ProcessFile] = 1
cgilua.prefunction[ cgilua.pre_Expression] = 1
cgilua.prefunction[ cgilua.pre_Code] = 1
cgilua.prefunction[ cgilua.pre_Loop ] = 1
cgilua.prefunction[ cgilua.pre_If ] = 1
cgilua.prefunction[ cgilua.pre_ProcessText] = 1
cgilua.prefunction[ cgilua.pre_IncludeText] = 1

cgilua.functionname = {}
cgilua.functionname[ cgilua.mkurl] = 'mkurl'
cgilua.functionname[ cgilua.relativeurl] = 'relativeurl'
cgilua.functionname[ cgilua.encodecgi] = 'encodecgi'
cgilua.functionname[ cgilua.encodetable] = 'encodetable'
cgilua.functionname[ cgilua.includehtml] = 'includehtml'
cgilua.functionname[ cgilua.preprocess] = 'preprocess'
cgilua.functionname[ cgilua.doscript] = 'doscript'
cgilua.functionname[ cgilua.splitpath] = 'splitpath'
cgilua.functionname[ cgilua.unescape] = 'unescape'
cgilua.functionname[ cgilua.escape] = 'escape'
cgilua.functionname[ cgilua.redirect] = 'redirect'
cgilua.functionname[ cgilua.htmlheader] = 'htmlheader'
cgilua.functionname[ cgilua.httpheader] = 'httpheader'
cgilua.functionname[ cgilua.loadlibrary] = 'loadlibrary'
cgilua.functionname[ cgilua.dofile] = 'dofile'
cgilua.functionname[ cgilua.tostring] = 'tostring'
cgilua.functionname[ cgilua.persistvar] = 'persistvar'
cgilua.functionname[ cgilua.savestate] = 'savestate'

