$debug

--=============================================================================

-- This file implements the preprocessing of a text file. Text not marked
-- to be processed is simply repeated on the output. There are three
-- kinds of marks that can be used in order to activate the processing:
--
-- (1) <!--$$ lua_code $$-->
--     where "lua_code" is a valid lua construction. This lua code is executed,
--     and if it contains expressions that write to the stdout, this
--     happens exactly at the point of the file where it is positioned.
--     Note that according to HTML syntax, everything enclosed by the
--     marks "<!--" and  "-->" is not displayed, so the lua code
--     is not visible on a browser if the file is viewed without
--     the preprocessing.
--
-- (2) $| lua_expression |$
--     where "lua_expression" is a valid lua expression. This lua 
--     expression is evaluated, and its result is printed. If the result 
--     is an element made with the "htmltoolkit" package, the element is 
--     recognized and properly printed. 
--
-- (3) <!--$$LOOP start='cmd_start', test='cond', action='cmd_action' $$-->
--       text
--     <!--$$ENDLOOP $$-->
--     where "cmd_start" is a lua code executed before the loop starts;
--     "cond" is a condition tested for the continuation of the loop,
--     if it is evaluated as false the loop stops; 
--     "cmd_action" is a lua code executed at the end of every loop cycle; and
--     "text" is a generic chunk of text that may contain any of the
--     three types of marks and that is continually submited for processing
--     until the loop stops.
--
--

--=============================================================================

_pre_LOOP      = format( '%c', 1 )
_pre_LOOPREM   = format( '%c', 2 )
_pre_ENDLOOP   = format( '%c', 15 )
_pre_IF        = format( '%c', 5 )
_pre_IFREM     = format( '%c', 6 )
_pre_ELSE      = format( '%c', 7 )
_pre_ENDIF     = format( '%c', 4 )
_pre_CODE      = format( '%c', 11 )
_pre_ENDCODE   = format( '%c', 14 )
_pre_EXP       = format( '%c', 3 )
_pre_ENDEXP    = format( '%c', 16 )

_pre_CLOSUREMARKS = _pre_LOOPREM.._pre_ENDLOOP.._pre_IFREM.._pre_ELSE.._pre_ENDIF.._pre_ENDCODE.._pre_ENDEXP
_pre_SPECIALMARKS = _pre_LOOP.._pre_IF.._pre_CODE.._pre_EXP.._pre_CLOSUREMARKS

cgilua.pre_userfile = nil
cgilua.pre_userline = 1
cgilua.pre_lastcount = 0

function cgilua.pre_CountLines( chunk, start, stop )
  local n = 0
  local s, e

  s, e = strfind( chunk, "\n", start )
  while s and s <= stop do
    n = n + 1
    s, e = strfind( chunk, "\n", e+1 )
  end
  cgilua.pre_userline = cgilua.pre_userline + n
  return n
end

function cgilua.pre_ReplaceTag( str, tagname, replopen, replclose, func )
  local t = str
   
  if not func then
    t = gsub( t, "<"..tagname..">(.-)</"..tagname..">", replopen.."%1"..replclose )
  elseif type( func ) == "function" then
    t = gsub( t, "<("..tagname..")>(.-)</"..tagname..">", func )
  else
    error( "<code>func field</code> is not a function." )
  end
 
  return t
end

function cgilua.pre_ProcessTags( str )
  local t = str
  
  if cgilua.custom_tags ~= nil and type( cgilua.custom_tags ) == "table" then
    local i = 1
    
    while cgilua.custom_tags[i] do
      local table = cgilua.custom_tags[i]
      t = cgilua.pre_ReplaceTag( t, table.tagname, table.open, table.close, table.func )
      i = i + 1
    end
  end
  return t
end

function cgilua.pre_PrepareText( text )
  local t

  t = cgilua.pre_ProcessTags( text )
  
  t = gsub( t, "<!%-%-%$%$%s*LOOP%s*(.-)%$%$%-%->", _pre_LOOP.."%1".._pre_LOOPREM )
  t = gsub( t,    "<!%-%-%$%$%s*ENDLOOP%s*%$%$%-%->",  _pre_ENDLOOP )
  t = gsub( t,    "<!%-%-%$%$%s*IF%s*(.-)%$%$%-%->",   _pre_IF.."%1".._pre_IFREM )
  t = gsub( t,    "<!%-%-%$%$%s*ELSE%s*%$%$%-%->",     _pre_ELSE )
  t = gsub( t,    "<!%-%-%$%$%s*ENDIF%s*%$%$%-%->",    _pre_ENDIF )
  t = gsub( t,    "<!%-%-%$%$",                        _pre_CODE )
  t = gsub( t,    "%$%$%-%->",                         _pre_ENDCODE )
  t = gsub( t,    "%$|",                               _pre_EXP )
  t = gsub( t,    "|%$",                               _pre_ENDEXP )

  return t
end


-- ProcessFile
-- This is the main function of the preprocessing package and the only one
-- that should be called outside this file. It receives an argument, the 
-- name of the file to be processed, and if there is no error opening
-- the file, the processing is started.

function cgilua.pre_ProcessFile( htmlfile )

  local oldin = _INPUT
  if readfrom( htmlfile ) then
    cgilua.pre_userfile = htmlfile
    cgilua.push( 'script', htmlfile )
    cgilua.pre_ProcessText( cgilua.pre_PrepareText( read( ".*" ) ) )
    cgilua.pop( 'script' )
    readfrom()
    _INPUT = oldin
    return 1
  else
    cgilua.push ( 'script', '' )
    cgilua.usererror( format( 'Error opening file <CODE>%s</CODE>. '..
                      'Preprocessing aborted.', htmlfile ) )
  end

  return nil
end

cgilua.includehtml = cgilua.pre_ProcessFile


------------------------------------------------------------------------------


-- Expression
-- This function evaluates expressions of the form "$|EXP|$" and
-- replaces them with the result of the lua expressions EXP.
-- If the expression is an element produced by the "htmltoolkit",
-- then the element itself knows how to be represented (the method
-- "Print" is called).

function cgilua.pre_Expression( expr )
  local repl

  -- evaluates the lua expression and print it
  repl = cgilua.douserstring( 'return '..expr, 'expression field', cgilua.pre_userline, cgilua.pre_userfile ) 

  if type( repl ) == 'table' and repl.tag then
    repl:Print()
  elseif type( repl ) == 'number' or type( repl ) == 'string' then
    write( repl )
  elseif repl ~= nil then
    cgilua.usererror( 'Error processing the expression '..
                      '<CODE>$|'..expr..'|$</CODE>. \nThe resulting '..
                      'item (type <CODE>'..type(repl)..'</CODE>) '..
                      'cannot be printed.')
  end
end

function cgilua.pre_Code( code )
  cgilua.douserstring( code, 'code field', cgilua.pre_userline, cgilua.pre_userfile )
end

------------------------------------------------------------------------------

-- Loop
-- This function executes a loop, using the control parameters specified
-- on the LOOP statement. The first argument is the chunk of text where
-- the corresponding ENDLOOP should be found. The second argument is
-- the control parameters obtained from the LOOP statement. 
-- A LOOP statement have the form:
-- <!--$$LOOP start='cmd', test='cond', action='cmd2' $$-->
-- Note that there must be 3 control parameters, separated
-- by commas: "start", "test" and "action". The value of this parameters
-- must be a string of a valid lua expression. 
-- "start" is a command executed one time before the loop starts;
-- "test" is a condition tested for the continuation of the loop; and
-- "action" is a command executed at the end of every loop cycle.

function cgilua.pre_Loop ( block )
  local control
  local s, e

  s, e, param = strfind( block, "(.-)".._pre_LOOPREM )
  if not s then cgilua.usererror( "missing <code>$$--&gt;</code> LOOP delimiter" ) end

  -- convert the control parameters into fields of a lua table
  control = cgilua.douserstring( 'return {'..param..'}', 'LOOP directive', cgilua.pre_userline, cgilua.pre_userfile )

  if not control.start then
    cgilua.usererror( "incorrect LOOP statement; missing <code>start</code> control field" )
  elseif not control.test then
    cgilua.usererror( "incorrect LOOP statement; missing <code>test</code> control field" )
  elseif not control.action then
    cgilua.usererror( "incorrect LOOP statement; missing <code>action</code> control field" )
  end

  -- do the "start" control statement
  cgilua.douserstring(control.start, 'LOOP directive (parameter start)', cgilua.pre_userline, cgilua.pre_userfile )

  -- while the "test" control statement is valid, run all over again
  -- the main PRE_ProcessText function, building a stack of chunks
  -- being processed. At the end, the "action" control statement
  -- is executed

  cgilua.pre_lastcount = cgilua.pre_userline
  while cgilua.douserstring('return '..control.test, 'LOOP directive (parameter test)', cgilua.pre_userline, cgilua.pre_userfile ) do
    cgilua.pre_ProcessText( block, e+1 )

    -- come back with the line counter for the next cycle
    cgilua.pre_userline = cgilua.pre_lastcount
    cgilua.douserstring(control.action, 'LOOP directive (parameter action)', cgilua.pre_userline, cgilua.pre_userfile )
  end
end

------------------------------------------------------------------------------

-- If
-- This function evaluates a condition and process the block of
-- text contained if.
-- A IF statement have the form:
-- <!--$$IF test='cond' $$-->
-- [ <!--$$ELSE $$--> ]
-- <!--$$ENDIF$$--> 
-- "test" is a condition tested for the continuation of the loop; and

function cgilua.pre_If ( block )

  local s, e, param          = strfind( block, "(.-)".._pre_IFREM )
  if not s then cgilua.usererror( "missing <code>$$--&gt;</code> IF delimiter" ) end

  local existelse, startelse = strfind( block, _pre_ELSE, e+1 )

  -- convert the control parameters into fields of a lua table
  local control = cgilua.douserstring( 'return {'..param..'}', 'IF directive', cgilua.pre_userline, cgilua.pre_userfile )

  if not control.test then
    cgilua.usererror( "incorrect IF statement; missing <code>test</code> control field" )
  end

  if cgilua.douserstring('return '..control.test, 'IF directive (parameter test)', cgilua.pre_userline, cgilua.pre_userfile ) then
    if existelse then 
      cgilua.pre_ProcessText( strsub( block, 1, startelse-1 ), e+1 )
    else
      cgilua.pre_CountLines( block, 1, e )
      cgilua.pre_ProcessText( block, e+1 )
    end
  elseif existelse then
    cgilua.pre_CountLines( block, 1, startelse )
    cgilua.pre_ProcessText( block, startelse+1 )
  end
end

------------------------------------------------------------------------------

-- ProcessText
-- This function coordinates the action of the others and executes the
-- main loop of the processing: it goes through the text searching
-- for special marks. There are two types of special marks:
-- 1) The pair  "<!--$$" and "$$-->";      
-- 2) The pair  "$|" and "|$".
-- On the first type all the text enclosed is captured and executed as
-- lua code. On the second type the text enclosed is considered a lua
-- expression and an attempt is made to print it. 

function cgilua.pre_ProcessText( text, start )

  if not start then 
    start = 1 
    cgilua.pre_userline = 1
  end

  local specialmark, _, marktype = strfind( text, "([".._pre_SPECIALMARKS.."])", start )

  if specialmark then

    write( strsub( text, start, specialmark-1 ) )
    cgilua.pre_CountLines( text, start, specialmark-1 )

    if marktype == _pre_EXP then

      local s, e, exp = strfind( text, "(%b".._pre_EXP.._pre_ENDEXP..")", specialmark )
      if not s then cgilua.usererror( "missing <code>|$</code> delimiter" ) end
      cgilua.pre_Expression( strsub( exp, 2, -2 ) )
      cgilua.pre_CountLines( text, s, e )
      cgilua.pre_ProcessText( text, e+1 )

    elseif marktype == _pre_CODE then

      local s, e, code = strfind( text, "(%b".._pre_CODE.._pre_ENDCODE..")", specialmark )
      if not s then cgilua.usererror( "missing <code>$$--&gt;</code> delimiter" ) end
      cgilua.pre_Code( strsub( code, 2, -2 ) )
      cgilua.pre_CountLines( text, s, e )
      cgilua.pre_ProcessText( text, e+1 )

    elseif marktype == _pre_LOOP then

      local s, e, loopblock = strfind( text, "(%b".._pre_LOOP.._pre_ENDLOOP..")", specialmark )
      if not s then cgilua.usererror( "missing <code>&lt;--$$ ENDLOOP $$--&gt;</code> delimiter" ) end
      cgilua.pre_Loop( strsub( loopblock, 2, -2 ) )
      cgilua.pre_CountLines( text, s, e )
      cgilua.pre_ProcessText( text, e+1 )

    elseif marktype == _pre_IF then

      local s, e, ifblock = strfind( text, "(%b".._pre_IF.._pre_ENDIF..")", specialmark )
      if not s then cgilua.usererror( "missing <code>&lt;--$$ ENDIF $$--&gt;</code> delimiter" ) end
      cgilua.pre_If( strsub( ifblock, 2, -2 ) )
      cgilua.pre_CountLines( text, s, e )
      cgilua.pre_ProcessText( text, e+1 )
      
    else

      notexpected = strfind( text, "([".._pre_CLOSUREMARKS.."])", start )
      if notexpected then
        cgilua.pre_CountLines( text, start, notexpected )
        cgilua.usererror( "missing opening preprocessor command" )
      end
  
    end
  else
    write( strsub( text, start ) )
  end
end

function cgilua.pre_IncludeText( text )
   cgilua.pre_ProcessText( cgilua.pre_PrepareText( text ) )  
end
------------------------------------------------------------------------------

cgilua.processtext = cgilua.pre_IncludeText
